import numpy
from scipy.interpolate import interp1d
from scipy.stats import truncnorm


class InfallDistributionGalacticus2024(object):
    """ACCRETION REDSHIFT PDF FROM GALACTICUS USING THE VERSION OF GALACTICUS AS OF FEB 2024 AND SELECTING ON
    SUBHALOS WITH A BOUND MASS > 10^6"""
    name = 'version2024'
    def __init__(self, z_lens):
        self._z_lens = z_lens
        self._counts = numpy.array([ 50,  93, 125, 180, 175, 144, 120, 117,  97,  82,  52,  51,  35,
        20,   9,   4,   4,   0,   1,   1])
        self._z_infall = numpy.array([ 0.53836189,  1.37376234,  2.2091628 ,  3.04456325,  3.87996371,
        4.71536416,  5.55076461,  6.38616507,  7.22156552,  8.05696598,
        8.89236643,  9.72776689, 10.56316734, 11.39856779, 12.23396825,
       13.0693687 , 13.90476916, 14.74016961, 15.57557007, 16.41097052]) - 0.5
        cdf = numpy.cumsum(self._counts)
        self._cdf = cdf / numpy.max(cdf)
        self._cdf_min = numpy.min(self._cdf)
        self._cdf_max = numpy.max(self._cdf)
        self._interp = interp1d(self._cdf, self._z_infall)

    def __call__(self, *args, **kwargs):
        u = numpy.random.uniform(self._cdf_min, self._cdf_max)
        z_infall = self._z_lens + self._interp(u)
        return z_infall

class InfallDistributionHybrid(object):
    name = 'hybrid'
    """Accretion redshift pdf that is a combination of directly infall and indirectly infall halos"""
    def __init__(self, z_lens, log_m_host):
        """

        :param z_lens: main deflector redshift
        :param log_m_host: log10 host halo mass in solar masses
        """
        self._m_host = 10**log_m_host
        self._z_lens = z_lens

    def __call__(self, m_sub):
        """
        Return the infall redshift for a subhalo with infall mass m_sub
        :param m_sub: infall mass in solar masses
        :return: infall redshift
        """
        mass_ratio = self._mass_ratio_in_bounds(m_sub / self._m_host)
        mu = self.z_inf_to_z_host_mean(mass_ratio)
        sig = self.z_inf_to_z_host_std(mass_ratio)
        bounds = [0.0, 15.0]
        z = float(truncnorm.rvs((bounds[0] - mu) / sig, (bounds[1] - mu) / sig,
                                 loc=mu, scale=sig))
        return self._z_lens + z

    @staticmethod
    def _mass_ratio_in_bounds(massRatio):
        """
        Make sure the ratio m_infall / m_host falls in the calibrated range 10^-5 - 10^-0.5
        :param massRatio:
        :return: the ratio m_infall / m_host
        """
        massRatio = max(10 ** -5.0, massRatio)
        massRatio = min(10 ** -0.5, massRatio)
        return massRatio

    @staticmethod
    def z_inf_to_z_host_mean(massRatio):
        """
        Return the mean infall redshift relative to the host redshift for subhalos with a mass ratio of
        massRatio = Msub / Mhost.
        :param massRatio: mass ratio of the subhao relative to the host, i.e. Msub / Mhost
        :return: the mean infall redshift relative to the host redshift
        """
        a = 3.35549949
        b = 3.20546995
        c = -2.91075587
        return a / (1 + b * (-numpy.log10(massRatio)) ** c)

    @staticmethod
    def z_inf_to_z_host_std(massRatio):
        """
        Return the standard deviation of infall redshift for subhalos with a mass ratio of
        massRatio = Msub / Mhost.
        :param massRatio: mass ratio of the subhao relative to the host, i.e. Msub / Mhost
        :return: the standard deviation of infall redshift
        """
        a = 1.97880156
        b = 4.17390702
        c = -2.14428416
        return a / (1 + b * (-numpy.log10(massRatio)) ** c)

class InfallDistributionDirect(object):
    name = 'direct'
    """Accretion redshift pdf meant for directly-infalling subhalos"""

    def __init__(self, z_lens, log_m_host):
        """

        :param z_lens: main deflector redshift
        :param log_m_host: log10 host halo mass in solar masses
        """
        self._m_host = 10 ** log_m_host
        self._z_lens = z_lens

    def __call__(self, m_sub):
        """
        Return the infall redshift for a subhalo with infall mass m_sub
        :param m_sub: infall mass in solar masses
        :return: infall redshift
        """
        mass_ratio = self._mass_ratio_in_bounds(m_sub / self._m_host)
        mu = self.z_inf_to_z_host_mean(mass_ratio)
        sig = self.z_inf_to_z_host_std(mass_ratio)
        bounds = [0.0, 15.0]
        z = float(truncnorm.rvs((bounds[0] - mu) / sig, (bounds[1] - mu) / sig,
                                 loc=mu, scale=sig))
        return self._z_lens + z

    @staticmethod
    def _mass_ratio_in_bounds(massRatio):
        """
        Make sure the ratio m_infall / m_host falls in the calibrated range 10^-5 - 10^-0.5
        :param massRatio:
        :return: the ratio m_infall / m_host
        """
        massRatio = max(10 ** -5.0, massRatio)
        massRatio = min(10 ** -0.5, massRatio)
        return massRatio

    @staticmethod
    def z_inf_to_z_host_mean(massRatio):
        """
        Return the mean infall redshift relative to the host redshift for subhalos with a mass ratio of
        massRatio = Msub / Mhost.
        :param massRatio: mass ratio of the subhao relative to the host, i.e. Msub / Mhost
        :return: the mean infall redshift relative to the host redshift
        """
        a = 3.3064042
        b = 4.528511
        c = -3.84346439
        return a / (1 + b * (-numpy.log10(massRatio)) ** c)

    @staticmethod
    def z_inf_to_z_host_std(massRatio):
        """
        Return the standard deviation of infall redshift for subhalos with a mass ratio of
        massRatio = Msub / Mhost.
        :param massRatio: mass ratio of the subhao relative to the host, i.e. Msub / Mhost
        :return: the standard deviation of infall redshift
        """
        a = 1.94929171
        b = 4.91504583
        c = -2.4187402
        return a / (1 + b * (-numpy.log10(massRatio)) ** c)

class InfallDistributionClusterDirect(object):
    name = 'direct_cluster'
    """Accretion redshift pdf meant for directly-infalling subhalos onto galaxy clusters M_host ~ 5*10^14
    Distribution is evaluated at 0.08 * R_vir

    M_host = 5.0e14 Msun
    mu: a = 2.05705664,  b = 9.43229827, c = -4.69408104
    sigma: a = 1.60922533,  b = 9.45120467, c = -2.5087969

    M_host = 1.0e15 Msun
    mu: a = 1.83256604,  b = 8.34655678, c = -4.26423697
    sigma: a = 1.8094271 ,  b = 9.17341074, c = -2.05254907
    """
    def __init__(self, z_lens, log_m_host):
        """

        :param z_lens: main deflector redshift
        """
        if log_m_host > 15.0 or log_m_host < 14.0:
            raise ValueError('infall redshift distribution for cluster-mass halos only valid for host halo masses '
                             'between 10^14 - 10^15 M_sun!')
        self._m_host = 10 ** log_m_host
        self._z_lens = z_lens

    def __call__(self, m_sub):
        """
        Return the infall redshift for a subhalo with infall mass m_sub
        :param m_sub: infall mass in solar masses
        :return: infall redshift
        """
        mass_ratio = self._mass_ratio_in_bounds(m_sub / self._m_host)
        mu = self.z_inf_to_z_host_mean(mass_ratio)
        sig = self.z_inf_to_z_host_std(mass_ratio)
        bounds = [0.0, 15.0]
        z = float(truncnorm.rvs((bounds[0] - mu) / sig, (bounds[1] - mu) / sig,
                                 loc=mu, scale=sig))
        return self._z_lens + z

    @staticmethod
    def _mass_ratio_in_bounds(massRatio):
        """
        Make sure the ratio m_infall / m_host falls in the calibrated range 10^-5 - 10^-0.5
        :param massRatio:
        :return: the ratio m_infall / m_host
        """
        massRatio = max(10 ** -5, massRatio)
        massRatio = min(10 ** -0.5, massRatio)
        return massRatio

    @staticmethod
    def z_inf_to_z_host_mean(massRatio):
        """
        Return the mean infall redshift relative to the host redshift for subhalos with a mass ratio of
        massRatio = Msub / Mhost.
        :param massRatio: mass ratio of the subhao relative to the host, i.e. Msub / Mhost
        :return: the mean infall redshift relative to the host redshift
        """
        a = 2.05705664
        b = 9.43229827
        c = -4.69408104
        return a / (1 + b * (-numpy.log10(massRatio)) ** c)

    @staticmethod
    def z_inf_to_z_host_std(massRatio):
        """
        Return the standard deviation of infall redshift for subhalos with a mass ratio of
        massRatio = Msub / Mhost.
        :param massRatio: mass ratio of the subhao relative to the host, i.e. Msub / Mhost
        :return: the standard deviation of infall redshift
        """
        a = 1.60922533
        b = 9.45120467
        c = -2.5087969
        return a / (1 + b * (-numpy.log10(massRatio)) ** c)



import numpy.testing as npt
from lenstronomy.Cosmo.lens_cosmo import LensCosmo as LensCosmoLenstronomy
from pyHalo.Halos.HaloModels.NFW_core import CoreNFWHalo
from pyHalo.Halos.concentration import ConcentrationConstant
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.Halos.HaloModels.TNFW import TNFWFieldHalo
import pytest
import numpy as np

class TestCNFWHalos(object):

    def setup_method(self):

        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.truncation_class = None
        self.c = 8
        self.concentration_class = ConcentrationConstant(self.lens_cosmo, self.c)
        self.lclenstronomy = LensCosmoLenstronomy(self.zhalo, self.zsource, astropy)

    def test_lenstronomy_params(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        kwargs_profile = {'beta': 0.99999}
        is_subhalo = False
        cnfw_halo = CoreNFWHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                      self.truncation_class, self.concentration_class, unique_tag)
        lenstronomy_params = cnfw_halo.lenstronomy_params[0][0]
        npt.assert_almost_equal(lenstronomy_params['r_core'], lenstronomy_params['Rs'], 3)

        profile_args = cnfw_halo.profile_args
        npt.assert_almost_equal(profile_args[0], self.c, 7)
        npt.assert_almost_equal(profile_args[1], kwargs_profile['beta'], 7)

    def test_mass3d(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        kwargs_profile = {'beta': 0.000001}
        is_subhalo = False
        zlens, zsource = 0.5, 2.0
        lens_cosmo = LensCosmo(zlens, zsource)
        cdm_halo = TNFWFieldHalo.simple_setup(m, 0.0, 0.0, zlens, tau=1000, lens_cosmo=lens_cosmo)
        cnfw_halo = CoreNFWHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                self.truncation_class, self.concentration_class, unique_tag)
        cnfw_halo._c = cdm_halo.c
        r = np.logspace(-0.5, 0.0, 2000) * cnfw_halo.nfw_params[1]
        rho_sidm = cnfw_halo.density_profile_3d(r)
        rho_cdm = cdm_halo.density_profile_3d(r)
        npt.assert_almost_equal(np.sum(rho_sidm / rho_cdm)/len(rho_sidm), 1, 1)


if __name__ == '__main__':
    pytest.main()




import numpy as np
from scipy.spatial.transform import Rotation
from pyHalo.Halos.HaloModels.TNFWFromParams import TNFWFromParams
from pyHalo.pyhalo import pyHalo
from pyHalo.Cosmology.geometry import Geometry
from pyHalo.single_realization import Realization
from pyHalo.Halos.galacticus_util.galacticus_util import GalacticusUtil
from pyHalo.Halos.galacticus_util.galacticus_filter import nodedata_filter_subhalos,nodedata_filter_tree,nodedata_filter_virialized,nodedata_filter_range,nodedata_apply_filter
import h5py


def DMFromGalacticus(galacticus_hdf5,z_source,cone_opening_angle_arcsec,tree_index,log_mlow_galacticus,log_mhigh_galacticus,
                     mass_range_is_bound=True, proj_angle_theta=0, proj_angle_phi=0,
                     nodedata_filter=None, galacticus_utilities=None, galacticus_params_additional=None,
                     galacticus_tabulate_tnfw_params=None, preset_model_los="CDM", **kwargs_los):
    """
    This generates a realization of halos using subhalo parameters provided from a specified tree in the galacticus file.
    See https://github.com/galacticusorg/galacticus/ for information on the galacticus galaxy formation model.

    :param galacticus_hdf5: Galacticus output hdf5 file. If str loads file from specified path
    :param z_source: source redshift
    :param cone_opening_angle_arcsec: The opening angle of the cone in arc seconds
    :param tree_index: The number of the tree to create a realization from.
        A single galacticus file contains multiple realizations (trees).
        NOTE: Trees output from galacticus are assigned a number starting at 1.
    :param log_galacticus_mlow: The log_10 of the minimum mass subhalo to include in the realization.
        Subhalos are filtered by bound or infall mass as set by setting mass_range_is_bound.
    :param log_galacticus_mhigh: The log_10 of the minimum subhalo to include in the realization
        Subhalos are filtered by bound or infall mass as set by setting mass_range_is_bound.
    :param mass_range_is_bound: If true subhalos are filtered bound mass, if false subhalos are filtered by infall mass.
    :param proj_angle_theta: Specifies the theta angle (in spherical coordinates) of the normal vector for the plane to project subhalo coordinates on.
        Should be in the range [0,pi]
    :param proj_angle_theta: Specifies the theta angle (in spherical coordinates) of the normal vector for the plane to project subhalo coordinates on.
        Should be in the range [0, 2 * pi]
    :param nodedata_filter: Expects a callable function that has input and output: (dict[str,np.ndarray], GalacticusUtil) -> np.ndarray[bool]
        ,subhalos are filtered based on the output np array. Defaults to None. If provided overrides default filtering.
    :param galacticus_params: Extra parameters to read when loading in galacticus hdf5 file.
    :param galacticus_tabulate_params: Expects a callable function that has input and output: (dict[str,np.ndarray], GalacticusUtil) -> dict[str,np.ndarray]
        Should return a dictionary containing a numpy array of args for use in creation of tnfw halos.
    :param preset_model_los: Specifies the preset model to use when generating the LOS subhalos. Defaults to CDM

    :return: A realization from Galacticus halos

    NOTE:
    For galacticus output files: All trees should output at the same redshift. The final output should include only one host halo per tree.
    An example galacticus output and it's parameter file can be found in example_notebooks/data
    """
    #Avoid circular import
    from pyHalo.preset_models import preset_model_from_name

    gutil = GalacticusUtil() if galacticus_utilities is None else galacticus_utilities

    MPC_TO_KPC = 1E3

    #Only read needed parameters to save memory.
    PARAMS_TO_READ_DEF = [
                            gutil.PARAM_X,
                            gutil.PARAM_Y,
                            gutil.PARAM_Z,
                            gutil.PARAM_TNFW_RHO_S,
                            gutil.PARAM_TNFW_RADIUS_TRUNCATION,
                            gutil.PARAM_RADIUS_VIRIAL,
                            gutil.PARAM_RADIUS_SCALE,
                            gutil.PARAM_MASS_BOUND,
                            gutil.PARAM_MASS_INFALL,
                            gutil.PARAM_ISOLATED,
                            gutil.PARAM_Z_LAST_ISOLATED,
                            gutil.PARAM_CONCENTRATION
                         ]

    if galacticus_params_additional is None:
        galacticus_params_additional = []
    else:
        galacticus_params_additional = list(galacticus_params_additional)

    params_to_read = galacticus_params_additional + PARAMS_TO_READ_DEF

    if isinstance(galacticus_hdf5,str):
        nodedata = gutil.read_nodedata_galacticus(galacticus_hdf5,params_to_read=params_to_read)
    elif isinstance(galacticus_hdf5, h5py.Group):
        nodedata = gutil.hdf5_read_galacticus_nodedata(galacticus_hdf5,params_to_read=params_to_read)
    else:
        nodedata = galacticus_hdf5

    #z_lens = np.mean(nodedata[gutil.PARAM_Z_LAST_ISOLATED][np.logical_not(nodedata_filter_subhalos(nodedata,gutil))])
    z_lens = 0.5

    tree_index = int(np.round(tree_index))

    # we create a realization of only line-of-sight halos by setting sigma_sub = 0.0
    kwargs_los['sigma_sub'] = 0.0
    kwargs_los["cone_opening_angle_arcsec"] = cone_opening_angle_arcsec
    kwargs_los["z_lens"] = z_lens
    kwargs_los["z_source"] = z_source

    halos_LOS = preset_model_from_name(preset_model_los)(**kwargs_los)

    # get lens_cosmo class from class containing LOS objects; note that this will work even if there are no LOS halos
    lens_cosmo = halos_LOS.lens_cosmo

    theta,phi = proj_angle_theta,proj_angle_phi

    # This rotation rotation maps the coordinates such that in the new coordinates zh = nh
    # Then we apply this rotation to x and y unit vectors so we get the x and y unit vectors in the plane
    rotation = Rotation.from_euler("zyz",(0,theta,phi))

    coords = np.asarray((nodedata[gutil.PARAM_X],nodedata[gutil.PARAM_Y],nodedata[gutil.PARAM_Z])) * MPC_TO_KPC

    xh_r = rotation.apply(np.array((1,0,0)))
    yh_r = rotation.apply(np.array((0,1,0)))

    kpc_per_arcsec_at_z = lens_cosmo.cosmo.kpc_proper_per_asec(z_lens)

    r2dmax_kpc = (cone_opening_angle_arcsec / 2) * kpc_per_arcsec_at_z

    coords_2d = np.asarray((np.dot(xh_r,coords),np.dot(yh_r,coords)))
    r2d_mag = np.linalg.norm(coords_2d,axis=0)

    filter_r2d = r2d_mag < r2dmax_kpc
    mass_key = gutil.PARAM_MASS_BOUND if mass_range_is_bound else gutil.PARAM_MASS_INFALL
    mass_range = 10.0**np.asarray((log_mlow_galacticus,log_mhigh_galacticus))

    # We should exclude nodes that are not valid subhalos, such as host halo nodes and nodes that are outside the virial radius.
    # (Galacticus is not calibrated for nodes outside of virial radius)
    if nodedata_filter is None:
        filter_subhalos = nodedata_filter_subhalos(nodedata,gutil)
        filter_virialized = nodedata_filter_virialized(nodedata,gutil)
        filter_mass = nodedata_filter_range(nodedata,mass_range,mass_key,gutil)
        filter_tree = nodedata_filter_tree(nodedata,tree_index,gutil)
        filter_combined = filter_subhalos & filter_virialized & filter_mass & filter_tree & filter_r2d

    else:
        filter_combined = nodedata_filter(nodedata,gutil)

    nodedata = nodedata_apply_filter(nodedata,filter_combined)
    coords_2d = coords_2d[:,filter_combined]
    r2d_mag = r2d_mag[filter_combined]
    coords = coords[:,filter_combined]
    r3d_mag = np.linalg.norm(coords,axis=0)

    def tabualate_params(nodedata,gutil):
        return {
            # The rhos_s factor of 4 comes from the this galacticus output is
            # The density normalization of the underlying NFW halo at r = rs
            # Multiply by 4 to get the normalization for the halo profile
            TNFWFromParams.KEY_RHO_S:       4 * nodedata[gutil.PARAM_TNFW_RHO_S] / (MPC_TO_KPC)**3,
            TNFWFromParams.KEY_RS :         nodedata[gutil.PARAM_RADIUS_SCALE] * MPC_TO_KPC,
            TNFWFromParams.KEY_RT :         nodedata[gutil.PARAM_TNFW_RADIUS_TRUNCATION] * MPC_TO_KPC,
            TNFWFromParams.KEY_RV :         nodedata[gutil.PARAM_CONCENTRATION] * nodedata[gutil.PARAM_RADIUS_SCALE] * MPC_TO_KPC,
            TNFWFromParams.KEY_Z_INFALL:    nodedata[gutil.PARAM_Z_LAST_ISOLATED],
            TNFWFromParams.KEY_ID :         nodedata[gutil.PARAM_NODE_ID]
        }

    galacticus_tabulate_tnfw_params = tabualate_params if galacticus_tabulate_tnfw_params is None else galacticus_tabulate_tnfw_params
    tnfw_args_all = galacticus_tabulate_tnfw_params(nodedata,gutil)
    halo_list = []
    for n,m_infall in enumerate(nodedata[gutil.PARAM_MASS_INFALL]):
        x,y = coords_2d[0][n], coords_2d[1][n]
        tnfw_args = {key:val[n] for key,val in tnfw_args_all.items()}
        halo_list.append(TNFWFromParams(m_infall,x,y,r3d_mag[n],z_lens,True,lens_cosmo,tnfw_args))
    subhalos_from_params = Realization.from_halos(halo_list,lens_cosmo,kwargs_halo_model={},
                                                    msheet_correction=False, rendering_classes=None)
    return halos_LOS.join(subhalos_from_params)

def DMSubhalosFromEmulator(z_lens, z_source, emulator_data_function, cone_opening_angle_arcsec=8.0):
    """
    Generate a pyHalo realization from an emulator trained on Galacticus simulations
    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param cone_opening_angle_arcsec: twice the radius in arcsec of the rendering area
    :param emulator_data_function: a callable function that on each call returns the properties of halos from the emulator
    :return: a pyHalo realization object
    """
    data = emulator_data_function()
    # FIRST WE CREATE AN INSTANCE OF PYHALO, WHICH SETS THE COSMOLOGY
    pyhalo = pyHalo(z_lens, z_source)
    # WE ALSO SPECIFY THE GEOMETRY OF THE RENDERING VOLUME
    geometry_type = 'DOUBLE_CONE'
    geometry = Geometry(pyhalo.cosmology, z_lens, z_source,
                        cone_opening_angle_arcsec, geometry_type)

    h = pyhalo.lens_cosmo.h
    MPC_TO_KPC = 1e3

    # Here, data = emulator_data from normalizing_flows.py
    massInfall = data[0]
    concentration = data[1]
    massBound = data[2]
    redshiftLastIsolated = data[3]
    lens_redshifts = np.array([z_lens] * len(massInfall))
    rt = data[4] # This has an MPC_TO_KPC conversion in the tabulate_params() function
    x_kpc = data[5] * MPC_TO_KPC
    y_kpc = data[6] * MPC_TO_KPC

    r_vir = ((3 * h * massInfall)/(4 * 200 * np.pi * pyhalo.lens_cosmo._nfw_param.rhoc_z(lens_redshifts)))**(1/3)
    r_vir = r_vir/h

    rs = r_vir/concentration
    ones = np.ones(len(concentration))
    rhos = h**2 * 200.0 / 3 * pyhalo.lens_cosmo._nfw_param.rhoc_z(lens_redshifts) * concentration**3 / (np.log(ones + concentration) - concentration / (ones + concentration))

    def tabulate_params():
        key_id = []
        for i in range(len(rhos)):
            key_id.append('TNFW')

        return {
            TNFWFromParams.KEY_RHO_S:       rhos / (MPC_TO_KPC)**3,
            TNFWFromParams.KEY_RS :         rs * MPC_TO_KPC,
            TNFWFromParams.KEY_RT :         rt * MPC_TO_KPC,
            TNFWFromParams.KEY_RV :         r_vir  * MPC_TO_KPC,
            TNFWFromParams.KEY_Z_INFALL:    redshiftLastIsolated,
            TNFWFromParams.KEY_ID :         key_id
        }
    tnfw_args_all = tabulate_params()
    halo_list = []
    for n,m_infall in enumerate(massInfall):
        tnfw_args = {key:val[n] for key,val in tnfw_args_all.items()}
        halo_list.append(TNFWFromParams(m_infall,x_kpc[n], y_kpc[n],None, z_lens,True,pyhalo.lens_cosmo,tnfw_args))

    subhalos_from_emulator = Realization.from_halos(halo_list, pyhalo.lens_cosmo, kwargs_halo_model= {},
                                                    msheet_correction=False, rendering_classes=None, geometry = geometry)

    return subhalos_from_emulator

def JointLOSWithEmulator(z_lens,
                           z_source,
                           emulator_data_function,
                           cone_opening_angle_arcsec,
                           preset_model_name_LOS,
                           **kwargs_preset_model):
    """
    Combine subhalos from an emulator with line-of-sight halos from another preset model
    :param z_lens: deflector redshift
    :param z_source: source redshift
    :param emulator_data_function: function for the emulator method
    :param cone_opening_angle_arcsec: twice the area of the rendering area in the main lens plane
    :param preset_model_name_LOS: the name of a preset model that generates line-of-sight halos
    :param kwargs_preset_model: keyword arguments for the line-of-sight preset model
    :return: a realization with subhalos generated from the emulator and line-of-sight halos generated by pyHalo
    """
    emulator_subhalos = DMSubhalosFromEmulator(z_lens, z_source, emulator_data_function, cone_opening_angle_arcsec)
    from pyHalo.preset_models import preset_model_from_name
    preset_model_function = preset_model_from_name(preset_model_name_LOS)
    los = preset_model_function(z_lens, z_source, **kwargs_preset_model)
    return emulator_subhalos.join(los)




from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.cnfw import CNFW as CNFWLenstronomy
import numpy as np

class CoreNFWHalo(Halo):
    # we use the pseudo nfw methods to normalize profile
    _pseudo_nfw = False
    """
    The main class for an NFW field halo profile without truncation

    See the base class in Halos/halo_base.py for the required routines for any instance of a Halo class
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag,
                 conserve_m200=True):

        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        self._cnfw_lenstronomy = CNFWLenstronomy()
        mdef = 'CNFW'
        super(CoreNFWHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                            lens_cosmo_instance, args, unique_tag)

    def density_profile_3d(self, r, profile_args=None, scaling=1.0):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        rho0, Rs, _ = self.nfw_params
        M0 = 4 * np.pi * rho0 * Rs ** 3
        r_core = self._args['beta'] * Rs
        return (M0 / 4 / np.pi) * ((r_core + r) * (r + Rs) ** 2) ** -1

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['CNFW']

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        rhos_kpc, rs_kpc, _ = self.nfw_params
        rhos_mpc = rhos_kpc * 1e9
        rs_mpc = rs_kpc * 1e-3
        Rs_angle, theta_Rs = self._lens_cosmo.nfw_physical2angle_fromNFWparams(rhos_mpc,
                                                                               rs_mpc,
                                                                               self.z)
        beta = self._args['beta']
        x, y = np.round(self.x, 4), np.round(self.y, 4)
        Rs_angle = np.round(Rs_angle, 10)
        theta_Rs = np.round(theta_Rs, 10)
        kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle, 'r_core': beta * Rs_angle,
                  'center_x': x, 'center_y': y}]
        return kwargs, None

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            concentration, beta = self.c, self._args['beta']
            self._profile_args = (concentration, beta)
        return self._profile_args
